package gas.world.blocks.defense.turrets;

import gas.entities.comp.*;
import mindustry.entities.*;
import gas.type.*;
import gas.world.blocks.logic.*;
import mindustry.content.*;
import gas.content.*;
import mindustry.world.blocks.defense.turrets.*;
import gas.world.blocks.payloads.*;
import mindustry.core.*;
import gas.world.meta.*;
import mindustry.game.EventType.*;
import gas.world.blocks.units.*;
import mindustry.world.blocks.heat.*;
import arc.audio.*;
import mindustry.world.blocks.legacy.*;
import mindustry.gen.*;
import mindustry.world.blocks.production.*;
import mindustry.world.draw.*;
import arc.math.*;
import mindustry.world.blocks.liquid.*;
import mindustry.entities.pattern.*;
import gas.world.blocks.heat.*;
import gas.world.blocks.defense.*;
import gas.world.draw.*;
import mindustry.entities.bullet.*;
import mindustry.world.blocks.distribution.*;
import gas.world.blocks.power.*;
import mindustry.world.*;
import gas.world.blocks.sandbox.*;
import mindustry.world.blocks.storage.*;
import gas.world.blocks.liquid.*;
import mindustry.game.*;
import gas.entities.*;
import mindustry.world.blocks.payloads.*;
import mindustry.world.blocks.campaign.*;
import gas.world.blocks.defense.turrets.*;
import gas.world.blocks.distribution.*;
import gas.world.*;
import mindustry.world.consumers.*;
import gas.world.blocks.gas.*;
import mindustry.world.meta.*;
import arc.math.geom.*;
import gas.world.blocks.campaign.*;
import mindustry.world.modules.*;
import gas.ui.*;
import mindustry.world.blocks.environment.*;
import gas.world.consumers.*;
import arc.graphics.g2d.*;
import mindustry.world.blocks.*;
import gas.world.blocks.production.GasGenericCrafter.*;
import arc.*;
import mindustry.world.blocks.logic.*;
import gas.world.modules.*;
import gas.world.blocks.*;
import arc.graphics.*;
import gas.*;
import gas.io.*;
import mindustry.ui.*;
import arc.func.*;
import gas.gen.*;
import mindustry.world.blocks.defense.turrets.Turret.*;
import gas.world.blocks.storage.*;
import mindustry.entities.Units.*;
import mindustry.world.blocks.units.*;
import mindustry.graphics.*;
import arc.util.*;
import arc.struct.*;
import arc.util.io.*;
import mindustry.world.blocks.defense.*;
import gas.world.blocks.production.*;
import gas.entities.bullets.*;
import mindustry.logic.*;
import mindustry.world.blocks.power.*;
import mindustry.type.*;
import mindustry.world.blocks.sandbox.*;
import static mindustry.Vars.*;

public class GasTurret extends GasReloadTurret {

    // after being logic-controlled and this amount of time passes, the turret will resume normal AI
    public final static float logicControlCooldown = 60 * 2;

    public final int timerTarget = timers++;

    /**
     * Ticks between attempt at finding a target.
     */
    public float targetInterval = 20;

    /**
     * Maximum ammo units stored.
     */
    public int maxAmmo = 30;

    /**
     * Ammo units used per shot.
     */
    public int ammoPerShot = 1;

    /**
     * If true, ammo is only consumed once per shot regardless of bullet count.
     */
    public boolean consumeAmmoOnce = false;

    /**
     * Minimum input heat required to fire.
     */
    public float heatRequirement = -1f;

    /**
     * Maximum efficiency possible, if this turret uses heat.
     */
    public float maxHeatEfficiency = 3f;

    /**
     * Bullet angle randomness in degrees.
     */
    public float inaccuracy = 0f;

    /**
     * Fraction of bullet velocity that is random.
     */
    public float velocityRnd = 0f;

    /**
     * Maximum angle difference in degrees at which turret will still try to shoot.
     */
    public float shootCone = 8f;

    /**
     * Turret shoot point.
     */
    public float shootX = 0f, shootY = Float.NEGATIVE_INFINITY;

    /**
     * Random spread on the X axis.
     */
    public float xRand = 0f;

    /**
     * Minimum bullet range. Used for artillery only.
     */
    public float minRange = 0f;

    /**
     * Minimum warmup needed to fire.
     */
    public float minWarmup = 0f;

    /**
     * If true, this turret will accurately target moving targets with respect to charge time.
     */
    public boolean accurateDelay = true;

    /**
     * If false, this turret can't move while charging.
     */
    public boolean moveWhileCharging = true;

    /**
     * pattern used for bullets
     */
    public ShootPattern shoot = new ShootPattern();

    /**
     * If true, this block targets air units.
     */
    public boolean targetAir = true;

    /**
     * If true, this block targets ground units and structures.
     */
    public boolean targetGround = true;

    /**
     * If true, this block targets friend blocks, to heal them.
     */
    public boolean targetHealing = false;

    /**
     * If true, this turret can be controlled by players.
     */
    public boolean playerControllable = true;

    /**
     * If true, this block will display ammo multipliers in its stats (irrelevant for certain types of turrets).
     */
    public boolean displayAmmoMultiplier = true;

    /**
     * Function for choosing which unit to target.
     */
    public Sortf unitSort = UnitSorts.closest;

    /**
     * Filter for types of units to attack.
     */
    public Boolf<Unit> unitFilter = u -> true;

    /**
     * Filter for types of buildings to attack.
     */
    public Boolf<Building> buildingFilter = b -> !b.block.underBullets;

    /**
     * Color of heat region drawn on top (if found)
     */
    public Color heatColor = Pal.turretHeat;

    /**
     * Optional override for all shoot effects.
     */
    @Nullable
    public Effect shootEffect;

    /**
     * Optional override for all smoke effects.
     */
    @Nullable
    public Effect smokeEffect;

    /**
     * Effect created when ammo is used. Not optional.
     */
    public Effect ammoUseEffect = Fx.none;

    /**
     * Sound emitted when a single bullet is shot.
     */
    public Sound shootSound = Sounds.shoot;

    /**
     * Sound emitted when shoot.firstShotDelay is >0 and shooting begins.
     */
    public Sound chargeSound = Sounds.none;

    /**
     * Range for pitch of shoot sound.
     */
    public float soundPitchMin = 0.9f, soundPitchMax = 1.1f;

    /**
     * Backwards Y offset of ammo eject effect.
     */
    public float ammoEjectBack = 1f;

    /**
     * Lerp speed of turret warmup.
     */
    public float shootWarmupSpeed = 0.1f;

    /**
     * If true, turret warmup is linear instead of a curve.
     */
    public boolean linearWarmup = false;

    /**
     * Visual amount by which the turret recoils back per shot.
     */
    public float recoil = 1f;

    /**
     * ticks taken for turret to return to starting position in ticks. uses reload time by default
     */
    public float recoilTime = -1f;

    /**
     * power curve applied to visual recoil
     */
    public float recoilPow = 1.8f;

    /**
     * ticks to cool down the heat region
     */
    public float cooldownTime = 20f;

    /**
     * Visual elevation of turret shadow, -1 to use defaults.
     */
    public float elevation = -1f;

    /**
     * How much the screen shakes per shot.
     */
    public float shake = 0f;

    /**
     * Defines drawing behavior for this turret.
     */
    public DrawBlock drawer = new DrawTurret();

    public GasTurret(String name) {
        super(name);
        liquidCapacity = 20f;
        quickRotate = false;
        outlinedIcon = 1;
    }

    @Override
    public boolean outputsItems() {
        return false;
    }

    @Override
    public void setStats() {
        super.setStats();
        stats.add(Stat.inaccuracy, (int) inaccuracy, StatUnit.degrees);
        stats.add(Stat.reload, 60f / (reload) * shoot.shots, StatUnit.perSecond);
        stats.add(Stat.targetsAir, targetAir);
        stats.add(Stat.targetsGround, targetGround);
        if (ammoPerShot != 1)
            stats.add(Stat.ammoUse, ammoPerShot, StatUnit.perShot);
        if (heatRequirement > 0)
            stats.add(Stat.input, heatRequirement, StatUnit.heatUnits);
    }

    @Override
    public void setBars() {
        super.setBars();
        if (heatRequirement > 0) {
            addBar("heat", (GasTurretBuild entity) -> new Bar(() -> Core.bundle.format("bar.heatpercent", (int) entity.heatReq, (int) (Math.min(entity.heatReq / heatRequirement, maxHeatEfficiency) * 100)), () -> Pal.lightOrange, () -> entity.heatReq / heatRequirement));
        }
    }

    @Override
    public void init() {
        if (shootY == Float.NEGATIVE_INFINITY)
            shootY = size * tilesize / 2f;
        if (elevation < 0)
            elevation = size / 2f;
        if (recoilTime < 0f)
            recoilTime = reload;
        if (cooldownTime < 0f)
            cooldownTime = reload;
        super.init();
    }

    @Override
    public void load() {
        super.load();
        drawer.load(this);
    }

    @Override
    public TextureRegion[] icons() {
        return drawer.finalIcons(this);
    }

    @Override
    public void getRegionsToOutline(Seq<TextureRegion> out) {
        drawer.getRegionsToOutline(this, out);
    }

    public class GasTurretBuild extends GasReloadTurretBuild implements ControlBlock {

        // TODO storing these as instance variables is horrible design
        /**
         * Turret sprite offset, based on recoil. Updated every frame.
         */
        public Vec2 recoilOffset = new Vec2();

        public Seq<AmmoEntry> ammo = new Seq<>();

        public int totalAmmo;

        public float curRecoil, heat, logicControlTime = -1;

        public float shootWarmup;

        public int totalShots;

        // turrets need to shoot once for 'visual reload' to be valid, otherwise they seem stuck at reload 0 when placed.
        public boolean visualReloadValid;

        public boolean logicShooting = false;

        @Nullable
        public Posc target;

        public Vec2 targetPos = new Vec2();

        public BlockUnitc unit = (BlockUnitc) UnitTypes.block.create(team);

        public boolean wasShooting;

        public int queuedBullets = 0;

        public float heatReq;

        public float[] sideHeat = new float[4];

        public float estimateDps() {
            if (!hasAmmo())
                return 0f;
            return shoot.shots / reload * 60f * (peekAmmo() == null ? 0f : peekAmmo().estimateDPS()) * potentialEfficiency * timeScale;
        }

        @Override
        public float range() {
            if (peekAmmo() != null) {
                return range + peekAmmo().rangeChange;
            }
            return range;
        }

        @Override
        public float warmup() {
            return shootWarmup;
        }

        @Override
        public float drawrot() {
            return rotation - 90;
        }

        @Override
        public boolean shouldConsume() {
            return isShooting() || reloadCounter < reload;
        }

        @Override
        public boolean canControl() {
            return playerControllable;
        }

        @Override
        public void control(LAccess type, double p1, double p2, double p3, double p4) {
            if (type == LAccess.shoot && !unit.isPlayer()) {
                targetPos.set(World.unconv((float) p1), World.unconv((float) p2));
                logicControlTime = logicControlCooldown;
                logicShooting = !Mathf.zero(p3);
            }
            super.control(type, p1, p2, p3, p4);
        }

        @Override
        public void control(LAccess type, Object p1, double p2, double p3, double p4) {
            if (type == LAccess.shootp && (unit == null || !unit.isPlayer())) {
                logicControlTime = logicControlCooldown;
                logicShooting = !Mathf.zero(p2);
                if (p1 instanceof Posc pos) {
                    targetPosition(pos);
                }
            }
            super.control(type, p1, p2, p3, p4);
        }

        @Override
        public double sense(LAccess sensor) {
            switch(sensor) {
                case ammo:
                    return totalAmmo;
                case ammoCapacity:
                    return maxAmmo;
                case rotation:
                    return rotation;
                case shootX:
                    return World.conv(targetPos.x);
                case shootY:
                    return World.conv(targetPos.y);
                case shooting:
                    return isShooting() ? 1 : 0;
                case progress:
                    return progress();
                default:
                    return super.sense(sensor);
            }
        }

        @Override
        public float progress() {
            return Mathf.clamp(reloadCounter / reload);
        }

        public boolean isShooting() {
            return (isControlled() ? unit.isShooting() : logicControlled() ? logicShooting : target != null);
        }

        @Override
        public Unit unit() {
            // make sure stats are correct
            unit.tile(this);
            unit.team(team);
            return (Unit) unit;
        }

        public boolean logicControlled() {
            return logicControlTime > 0;
        }

        public boolean isActive() {
            return (target != null || wasShooting) && enabled;
        }

        public void targetPosition(Posc pos) {
            if (!hasAmmo() || pos == null)
                return;
            BulletType bullet = peekAmmo();
            var offset = Tmp.v1.setZero();
            // when delay is accurate, assume unit has moved by chargeTime already
            if (accurateDelay && pos instanceof Hitboxc h) {
                offset.set(h.deltaX(), h.deltaY()).scl(shoot.firstShotDelay / Time.delta);
            }
            targetPos.set(Predict.intercept(this, pos, offset.x, offset.y, bullet.speed <= 0.01f ? 99999999f : bullet.speed));
            if (targetPos.isZero()) {
                targetPos.set(pos);
            }
        }

        @Override
        public void draw() {
            drawer.draw(this);
        }

        @Override
        public void updateTile() {
            if (!validateTarget())
                target = null;
            float warmupTarget = isShooting() && canConsume() ? 1f : 0f;
            if (linearWarmup) {
                shootWarmup = Mathf.approachDelta(shootWarmup, warmupTarget, shootWarmupSpeed);
            } else {
                shootWarmup = Mathf.lerpDelta(shootWarmup, warmupTarget, shootWarmupSpeed);
            }
            wasShooting = false;
            curRecoil = Math.max(curRecoil - Time.delta / recoilTime, 0);
            heat = Math.max(heat - Time.delta / cooldownTime, 0);
            unit.tile(this);
            unit.rotation(rotation);
            unit.team(team);
            recoilOffset.trns(rotation, -Mathf.pow(curRecoil, recoilPow) * recoil);
            if (logicControlTime > 0) {
                logicControlTime -= Time.delta;
            }
            if (heatRequirement > 0) {
                heatReq = calculateHeat(sideHeat);
            }
            // turret always reloads regardless of whether it's targeting something
            updateReload();
            if (hasAmmo()) {
                if (Float.isNaN(reloadCounter))
                    reloadCounter = 0;
                if (timer(timerTarget, targetInterval)) {
                    findTarget();
                }
                if (validateTarget()) {
                    boolean canShoot = true;
                    if (isControlled()) {
                        // player behavior
                        targetPos.set(unit.aimX(), unit.aimY());
                        canShoot = unit.isShooting();
                    } else if (logicControlled()) {
                        // logic behavior
                        canShoot = logicShooting;
                    } else {
                        // default AI behavior
                        targetPosition(target);
                        if (Float.isNaN(rotation))
                            rotation = 0;
                    }
                    float targetRot = angleTo(targetPos);
                    if (shouldTurn()) {
                        turnToTarget(targetRot);
                    }
                    if (Angles.angleDist(rotation, targetRot) < shootCone && canShoot) {
                        wasShooting = true;
                        updateShooting();
                    }
                }
            }
            if (coolant != null) {
                updateCooling();
            }
        }

        @Override
        public void handleLiquid(Building source, Liquid liquid, float amount) {
            if (coolant != null && liquids.currentAmount() <= 0.001f) {
                Events.fire(Trigger.turretCool);
            }
            super.handleLiquid(source, liquid, amount);
        }

        protected boolean validateTarget() {
            return !Units.invalidateTarget(target, canHeal() ? Team.derelict : team, x, y) || isControlled() || logicControlled();
        }

        protected boolean canHeal() {
            return targetHealing && hasAmmo() && peekAmmo().collidesTeam && peekAmmo().heals();
        }

        protected void findTarget() {
            float range = range();
            if (targetAir && !targetGround) {
                target = Units.bestEnemy(team, x, y, range, e -> !e.dead() && !e.isGrounded() && unitFilter.get(e), unitSort);
            } else {
                target = Units.bestTarget(team, x, y, range, e -> !e.dead() && unitFilter.get(e) && (e.isGrounded() || targetAir) && (!e.isGrounded() || targetGround), b -> targetGround && buildingFilter.get(b), unitSort);
                if (target == null && canHeal()) {
                    target = Units.findAllyTile(team, x, y, range, b -> b.damaged() && b != this);
                }
            }
        }

        protected void turnToTarget(float targetRot) {
            rotation = Angles.moveToward(rotation, targetRot, rotateSpeed * delta() * potentialEfficiency);
        }

        public boolean shouldTurn() {
            return moveWhileCharging || !charging();
        }

        @Override
        public void updateEfficiencyMultiplier() {
            if (heatRequirement > 0) {
                efficiency *= Math.min(heatReq / heatRequirement, maxHeatEfficiency);
            }
        }

        /**
         * Consume ammo and return a type.
         */
        public BulletType useAmmo() {
            if (cheating())
                return peekAmmo();
            AmmoEntry entry = ammo.peek();
            entry.amount -= ammoPerShot;
            if (entry.amount <= 0)
                ammo.pop();
            totalAmmo -= ammoPerShot;
            totalAmmo = Math.max(totalAmmo, 0);
            return entry.type();
        }

        /**
         * @return the ammo type that will be returned if useAmmo is called.
         */
        @Nullable
        public BulletType peekAmmo() {
            return ammo.size == 0 ? null : ammo.peek().type();
        }

        /**
         * @return whether the turret has ammo.
         */
        public boolean hasAmmo() {
            // used for "side-ammo" like gas in some turrets
            if (!canConsume())
                return false;
            // skip first entry if it has less than the required amount of ammo
            if (ammo.size >= 2 && ammo.peek().amount < ammoPerShot && ammo.get(ammo.size - 2).amount >= ammoPerShot) {
                totalAmmo -= ammo.pop().amount;
            }
            return ammo.size > 0 && ammo.peek().amount >= ammoPerShot;
        }

        public boolean charging() {
            return queuedBullets > 0 && shoot.firstShotDelay > 0;
        }

        protected void updateReload() {
            float multiplier = hasAmmo() ? peekAmmo().reloadMultiplier : 1f;
            reloadCounter += delta() * multiplier * baseReloadSpeed();
            // cap reload for visual reasons
            reloadCounter = Math.min(reloadCounter, reload);
        }

        protected void updateShooting() {
            if (reloadCounter >= reload && !charging() && shootWarmup >= minWarmup) {
                visualReloadValid = true;
                BulletType type = peekAmmo();
                shoot(type);
                reloadCounter %= reload;
            }
        }

        protected void shoot(BulletType type) {
            float bulletX = x + Angles.trnsx(rotation - 90, shootX, shootY), bulletY = y + Angles.trnsy(rotation - 90, shootX, shootY);
            if (shoot.firstShotDelay > 0) {
                chargeSound.at(bulletX, bulletY, Mathf.random(soundPitchMin, soundPitchMax));
                type.chargeEffect.at(bulletX, bulletY, rotation);
            }
            shoot.shoot(totalShots, (xOffset, yOffset, angle, delay, mover) -> {
                queuedBullets++;
                if (delay > 0f) {
                    Time.run(delay, () -> bullet(type, xOffset, yOffset, angle, mover));
                } else {
                    bullet(type, xOffset, yOffset, angle, mover);
                }
                totalShots++;
            });
            if (consumeAmmoOnce) {
                useAmmo();
            }
        }

        protected void bullet(BulletType type, float xOffset, float yOffset, float angleOffset, Mover mover) {
            queuedBullets--;
            if (dead || (!consumeAmmoOnce && !hasAmmo()))
                return;
            float xSpread = Mathf.range(xRand), bulletX = x + Angles.trnsx(rotation - 90, shootX + xOffset + xSpread, shootY + yOffset), bulletY = y + Angles.trnsy(rotation - 90, shootX + xOffset + xSpread, shootY + yOffset), shootAngle = rotation + angleOffset + Mathf.range(inaccuracy);
            float lifeScl = type.scaleLife ? Mathf.clamp(Mathf.dst(bulletX, bulletY, targetPos.x, targetPos.y) / type.range, minRange / type.range, range() / type.range) : 1f;
            // TODO aimX / aimY for multi shot turrets?
            handleBullet(type.create(this, team, bulletX, bulletY, shootAngle, -1f, (1f - velocityRnd) + Mathf.random(velocityRnd), lifeScl, null, mover, targetPos.x, targetPos.y), xOffset, yOffset, shootAngle - rotation);
            (shootEffect == null ? type.shootEffect : shootEffect).at(bulletX, bulletY, rotation + angleOffset, type.hitColor);
            (smokeEffect == null ? type.smokeEffect : smokeEffect).at(bulletX, bulletY, rotation + angleOffset, type.hitColor);
            shootSound.at(bulletX, bulletY, Mathf.random(soundPitchMin, soundPitchMax));
            ammoUseEffect.at(x - Angles.trnsx(rotation, ammoEjectBack), y - Angles.trnsy(rotation, ammoEjectBack), rotation * Mathf.sign(xOffset));
            if (shake > 0) {
                Effect.shake(shake, shake, this);
            }
            curRecoil = 1f;
            heat = 1f;
            if (!consumeAmmoOnce) {
                useAmmo();
            }
        }

        protected void handleBullet(@Nullable Bullet bullet, float offsetX, float offsetY, float angleOffset) {
        }

        @Override
        public void write(Writes write) {
            super.write(write);
            write.f(reloadCounter);
            write.f(rotation);
        }

        @Override
        public void read(Reads read, byte revision) {
            super.read(read, revision);
            if (revision >= 1) {
                reloadCounter = read.f();
                rotation = read.f();
            }
        }

        @Override
        public byte version() {
            return 1;
        }
    }
}
