package gas.world.blocks.units;

import mindustry.entities.units.*;
import gas.world.blocks.distribution.*;
import mindustry.entities.*;
import gas.type.*;
import gas.world.blocks.campaign.*;
import gas.world.blocks.logic.*;
import mindustry.content.*;
import gas.content.*;
import mindustry.world.blocks.defense.turrets.*;
import gas.world.blocks.payloads.*;
import gas.io.*;
import gas.world.meta.*;
import mindustry.game.EventType.*;
import gas.world.blocks.units.*;
import mindustry.world.blocks.heat.*;
import arc.util.*;
import mindustry.world.blocks.legacy.*;
import mindustry.gen.*;
import mindustry.world.blocks.production.*;
import mindustry.world.draw.*;
import arc.math.*;
import mindustry.world.blocks.liquid.*;
import mindustry.world.meta.*;
import gas.world.blocks.heat.*;
import gas.world.blocks.defense.*;
import mindustry.world.blocks.distribution.*;
import gas.world.blocks.power.*;
import mindustry.world.*;
import gas.world.blocks.sandbox.*;
import mindustry.world.blocks.storage.*;
import gas.world.blocks.liquid.*;
import gas.entities.*;
import mindustry.world.blocks.campaign.*;
import gas.world.blocks.defense.turrets.*;
import mindustry.ui.*;
import gas.world.*;
import mindustry.world.consumers.*;
import mindustry.world.blocks.units.Reconstructor.*;
import gas.world.blocks.gas.*;
import mindustry.io.*;
import arc.math.geom.*;
import gas.entities.comp.*;
import mindustry.world.modules.*;
import gas.ui.*;
import mindustry.world.blocks.environment.*;
import mindustry.*;
import gas.world.consumers.*;
import mindustry.world.blocks.payloads.*;
import mindustry.world.blocks.*;
import gas.world.blocks.production.GasGenericCrafter.*;
import arc.*;
import mindustry.world.blocks.logic.*;
import gas.world.modules.*;
import gas.world.blocks.*;
import gas.*;
import arc.graphics.g2d.*;
import gas.world.draw.*;
import gas.gen.*;
import gas.world.blocks.storage.*;
import mindustry.world.blocks.units.*;
import mindustry.graphics.*;
import gas.world.blocks.production.*;
import arc.struct.*;
import arc.util.io.*;
import mindustry.world.blocks.defense.*;
import gas.entities.bullets.*;
import mindustry.logic.*;
import mindustry.world.blocks.power.*;
import mindustry.type.*;
import mindustry.world.blocks.sandbox.*;
import static mindustry.Vars.*;

public class GasReconstructor extends GasUnitBlock {

    public float constructTime = 60 * 2;

    public Seq<UnitType[]> upgrades = new Seq<>();

    public int[] capacities = {};

    public GasReconstructor(String name) {
        super(name);
        regionRotated1 = 1;
        regionRotated2 = 2;
        commandable = true;
    }

    @Override
    public void drawPlanRegion(BuildPlan plan, Eachable<BuildPlan> list) {
        Draw.rect(region, plan.drawx(), plan.drawy());
        Draw.rect(inRegion, plan.drawx(), plan.drawy(), plan.rotation * 90);
        Draw.rect(outRegion, plan.drawx(), plan.drawy(), plan.rotation * 90);
        Draw.rect(topRegion, plan.drawx(), plan.drawy());
    }

    @Override
    public TextureRegion[] icons() {
        return new TextureRegion[] { region, inRegion, outRegion, topRegion };
    }

    @Override
    public void setBars() {
        super.setBars();
        addBar("progress", (GasReconstructorBuild entity) -> new Bar("bar.progress", Pal.ammo, entity::fraction));
        addBar("units", (GasReconstructorBuild e) -> new Bar(() -> e.unit() == null ? "[lightgray]" + Iconc.cancel : Core.bundle.format("bar.unitcap", Fonts.getUnicodeStr(e.unit().name), e.team.data().countType(e.unit()), Units.getCap(e.team)), () -> Pal.power, () -> e.unit() == null ? 0f : (float) e.team.data().countType(e.unit()) / Units.getCap(e.team)));
    }

    @Override
    public void setStats() {
        stats.timePeriod = constructTime;
        super.setStats();
        stats.add(Stat.productionTime, constructTime / 60f, StatUnit.seconds);
        stats.add(Stat.output, table -> {
            table.row();
            for (var upgrade : upgrades) {
                float size = 8 * 3;
                if (upgrade[0].unlockedNow() && upgrade[1].unlockedNow()) {
                    table.image(upgrade[0].uiIcon).size(size).padRight(4).padLeft(10).scaling(Scaling.fit).right();
                    table.add(upgrade[0].localizedName).left();
                    table.add("[lightgray] -> ");
                    table.image(upgrade[1].uiIcon).size(size).padRight(4).scaling(Scaling.fit);
                    table.add(upgrade[1].localizedName).left();
                    table.row();
                }
            }
        });
    }

    @Override
    public void init() {
        capacities = new int[Vars.content.items().size];
        ConsumeItems cons = findConsumer(c -> c instanceof ConsumeItems);
        if (cons != null) {
            for (var stack : cons.items) {
                capacities[stack.item.id] = Math.max(capacities[stack.item.id], stack.amount * 2);
                itemCapacity = Math.max(itemCapacity, stack.amount * 2);
            }
        }
        super.init();
    }

    public void addUpgrade(UnitType from, UnitType to) {
        upgrades.add(new UnitType[] { from, to });
    }

    public class GasReconstructorBuild extends GasUnitBuild {

        @Nullable
        public Vec2 commandPos;

        public float fraction() {
            return progress / constructTime;
        }

        @Override
        public Vec2 getCommandPosition() {
            return commandPos;
        }

        @Override
        public void onCommand(Vec2 target) {
            commandPos = target;
        }

        @Override
        public boolean acceptUnitPayload(Unit unit) {
            return hasUpgrade(unit.type) && !upgrade(unit.type).isBanned();
        }

        @Override
        public boolean acceptPayload(Building source, Payload payload) {
            if (!(this.payload == null && (this.enabled || source == this) && relativeTo(source) != rotation && payload instanceof UnitPayload pay)) {
                return false;
            }
            var upgrade = upgrade(pay.unit.type);
            if (upgrade != null) {
                if (!upgrade.unlockedNowHost() && !team.isAI()) {
                    // flash "not researched"
                    pay.showOverlay(Icon.tree);
                }
                if (upgrade.isBanned()) {
                    // flash an X, meaning 'banned'
                    pay.showOverlay(Icon.cancel);
                }
            }
            return upgrade != null && (team.isAI() || upgrade.unlockedNowHost()) && !upgrade.isBanned();
        }

        @Override
        public int getMaximumAccepted(Item item) {
            return capacities[item.id];
        }

        @Override
        public void overwrote(Seq<Building> builds) {
            if (builds.first().block == block) {
                items.add(builds.first().items);
            }
        }

        @Override
        public void draw() {
            Draw.rect(region, x, y);
            // draw input
            boolean fallback = true;
            for (int i = 0; i < 4; i++) {
                if (blends(i) && i != rotation) {
                    Draw.rect(inRegion, x, y, (i * 90) - 180);
                    fallback = false;
                }
            }
            if (fallback)
                Draw.rect(inRegion, x, y, rotation * 90);
            Draw.rect(outRegion, x, y, rotdeg());
            if (constructing() && hasArrived()) {
                Draw.draw(Layer.blockOver, () -> {
                    Draw.alpha(1f - progress / constructTime);
                    Draw.rect(payload.unit.type.fullIcon, x, y, payload.rotation() - 90);
                    Draw.reset();
                    Drawf.construct(this, upgrade(payload.unit.type), payload.rotation() - 90f, progress / constructTime, speedScl, time);
                });
            } else {
                Draw.z(Layer.blockOver);
                drawPayload();
            }
            Draw.z(Layer.blockOver + 0.1f);
            Draw.rect(topRegion, x, y);
        }

        @Override
        public Object senseObject(LAccess sensor) {
            if (sensor == LAccess.config)
                return unit();
            return super.senseObject(sensor);
        }

        @Override
        public void updateTile() {
            boolean valid = false;
            if (payload != null) {
                // check if offloading
                if (!hasUpgrade(payload.unit.type)) {
                    moveOutPayload();
                } else {
                    // update progress
                    if (moveInPayload()) {
                        if (efficiency > 0) {
                            valid = true;
                            progress += edelta() * state.rules.unitBuildSpeed(team);
                        }
                        // upgrade the unit
                        if (progress >= constructTime) {
                            payload.unit = upgrade(payload.unit.type).create(payload.unit.team());
                            if (commandPos != null && payload.unit.isCommandable()) {
                                payload.unit.command().commandPosition(commandPos);
                            }
                            progress %= 1f;
                            Effect.shake(2f, 3f, this);
                            Fx.producesmoke.at(this);
                            consume();
                            Events.fire(new UnitCreateEvent(payload.unit, this));
                        }
                    }
                }
            }
            speedScl = Mathf.lerpDelta(speedScl, Mathf.num(valid), 0.05f);
            time += edelta() * speedScl * state.rules.unitBuildSpeed(team);
        }

        @Override
        public double sense(LAccess sensor) {
            if (sensor == LAccess.progress)
                return Mathf.clamp(fraction());
            return super.sense(sensor);
        }

        @Override
        public boolean shouldConsume() {
            return constructing() && enabled;
        }

        public UnitType unit() {
            if (payload == null)
                return null;
            UnitType t = upgrade(payload.unit.type);
            return t != null && (t.unlockedNowHost() || team.isAI()) ? t : null;
        }

        public boolean constructing() {
            return payload != null && hasUpgrade(payload.unit.type);
        }

        public boolean hasUpgrade(UnitType type) {
            UnitType t = upgrade(type);
            return t != null && (t.unlockedNowHost() || team.isAI()) && !type.isBanned();
        }

        public UnitType upgrade(UnitType type) {
            UnitType[] r = upgrades.find(u -> u[0] == type);
            return r == null ? null : r[1];
        }

        @Override
        public byte version() {
            return 2;
        }

        @Override
        public void write(Writes write) {
            super.write(write);
            write.f(progress);
            TypeIO.writeVecNullable(write, commandPos);
        }

        @Override
        public void read(Reads read, byte revision) {
            super.read(read, revision);
            if (revision >= 1) {
                progress = read.f();
            }
            if (revision >= 2) {
                commandPos = TypeIO.readVecNullable(read);
            }
        }
    }
}
